/* Copyright (c) 2014 Nordic Semiconductor. All Rights Reserved.
 *
 * The information contained herein is property of Nordic Semiconductor ASA.
 * Terms and conditions of usage are described in detail in NORDIC
 * SEMICONDUCTOR STANDARD SOFTWARE LICENSE AGREEMENT.
 *
 * Licensees are granted free, non-transferable use of the information. NO
 * WARRANTY of ANY KIND is provided. This heading must NOT be removed from
 * the file.
 *
 */

#include <stdbool.h>
#include <stdint.h>
#include <string.h>
#include "app_scheduler.h"
#include "app_util.h"
#include "nrf.h"
#include "boards.h"
#include "micro_esb.h"
#include "uesb_error_codes.h"
#include "nrf_gpio.h"
#include "simple_uart.h"
#include "fifo.h"

#define APP_SCHED_QUEUE_SIZE 10
#define APP_SCHED_EVENT_SIZE sizeof(uesb_payload_t)

#define UESB_PIPE0_ADDR {0xD2, 0xF0, 0xF0, 0xF0, 0xF0}
#define UESB_PIPE1_ADDR {0xE1, 0xF0, 0xF0, 0xF0, 0xF0}
#define UESB_PIPE2_ADDR 0x66

#define UESB_TX_PIPE 0

static fifo_t s_uart_fifo;

// Required by app_scheduler
void critical_region_enter(void)
{
    __disable_irq();
}

// Required by app_scheduler
void critical_region_exit(void)
{
    __enable_irq();
}

static void uart_print_evt_handler(void * p_event_data, uint16_t event_size)
{
    uint8_t * char_ptr;
    
    APP_ERROR_CHECK_BOOL(event_size != 0);
    
    char_ptr = (uint8_t *) p_event_data;
    
    for (int i = 0; i < event_size; ++i)
    {
        simple_uart_put(*char_ptr++);
    }
}

static void uesb_send_pkt_evt_handler(void * p_event_data, uint16_t event_size)
{
    uesb_payload_t payload;
    
    APP_ERROR_CHECK_BOOL(event_size <= UESB_CORE_MAX_PAYLOAD_LENGTH);
    
    payload.pipe  = UESB_TX_PIPE;
    payload.length = event_size;
    memcpy(payload.data, p_event_data, event_size);
    
    // Stop RX
    uesb_stop_rx(); // Ignore possible UESB_ERROR_NOT_IN_RX_MODE return code
    
    uesb_write_tx_payload(&payload); // Ignore possible errors
}

static void uesb_start_rx_evt_handler(void * p_event_data, uint16_t event_size)
{
    uint32_t err_code;
    
    err_code = uesb_start_rx();
    if (err_code != UESB_SUCCESS)
    {
        // Retry operation
        app_sched_event_put(0, 0, uesb_start_rx_evt_handler);
    }
}  

void uesb_event_handler()
{
    static uint32_t rf_interrupts;
    static uint32_t tx_attempts; // Retries attempted by ESB
    
    uesb_get_clear_interrupts(&rf_interrupts);
    
    if(rf_interrupts & UESB_INT_TX_SUCCESS_MSK)
    {   
        uint32_t err_code;
        
        // Data successfully sent. Start RX again
        err_code = app_sched_event_put(0, 0, uesb_start_rx_evt_handler);
        APP_ERROR_CHECK(err_code);
        nrf_gpio_pin_toggle(21);
    }
    
    if(rf_interrupts & UESB_INT_TX_FAILED_MSK)
    {
        // Data failed to be sent: TODO: try again later
        uesb_flush_tx();
        nrf_gpio_pin_toggle(22);
    }
    
    if(rf_interrupts & UESB_INT_RX_DR_MSK)
    {
        uint32_t       err_code;
        uesb_payload_t payload;
        
        // Data has been received: schedule printout
        uesb_read_rx_payload(&payload);
        
        err_code = app_sched_event_put(payload.data, payload.length, uart_print_evt_handler);
        APP_ERROR_CHECK(err_code);
        nrf_gpio_pin_toggle(23);
    }
    
    uesb_get_tx_attempts(&tx_attempts);
}

static void uesb_config(void)
{
    uint8_t rx_addr_p0[] = UESB_PIPE0_ADDR;
    uint8_t rx_addr_p1[] = UESB_PIPE1_ADDR;
    uint8_t rx_addr_p2   = UESB_PIPE2_ADDR;
    
    uesb_config_t uesb_config       = UESB_DEFAULT_CONFIG;
    uesb_config.rf_channel          = 5;
    uesb_config.crc                 = UESB_CRC_16BIT;
    uesb_config.retransmit_count    = 6;
    uesb_config.retransmit_delay    = 500;
    uesb_config.dynamic_ack_enabled = true;
    uesb_config.protocol            = UESB_PROTOCOL_ESB_DPL;
    uesb_config.bitrate             = UESB_BITRATE_1MBPS;
    uesb_config.event_handler       = uesb_event_handler;

    uesb_init(&uesb_config);

    uesb_set_address(UESB_ADDRESS_PIPE0, rx_addr_p0);
    uesb_set_address(UESB_ADDRESS_PIPE1, rx_addr_p1);
    uesb_set_address(UESB_ADDRESS_PIPE2, &rx_addr_p2);
}

void UART0_IRQHandler(void)
{
    if (NRF_UART0->EVENTS_RXDRDY != 0)
    {
        uint8_t rxd_char;
        
        NRF_UART0->EVENTS_RXDRDY = 0;
        
        rxd_char = (uint8_t) NRF_UART0->RXD;
        
        fifo_put_char(&s_uart_fifo, rxd_char);
        
        // Send data packet when newline character is detected or when packet size has been reached
        if (rxd_char == '\n' || fifo_num_elem_get(&s_uart_fifo) >= UESB_CORE_MAX_PAYLOAD_LENGTH)
        {
            uint8_t  char_buf[UESB_CORE_MAX_PAYLOAD_LENGTH];
            uint32_t char_buf_len;
            uint32_t err_code;
            
            char_buf_len = sizeof(char_buf);
            fifo_get_pkt(&s_uart_fifo, char_buf, &char_buf_len);
            
            err_code = app_sched_event_put(char_buf, char_buf_len, uesb_send_pkt_evt_handler);
            APP_ERROR_CHECK(err_code);
        }
    }
}

static void uart_config(void)
{
    simple_uart_config(RTS_PIN_NUMBER, TX_PIN_NUMBER, CTS_PIN_NUMBER, RX_PIN_NUMBER, false);
    // Enable interrupt on UART RX
    NVIC_ClearPendingIRQ(UART0_IRQn);
    NVIC_SetPriority(UART0_IRQn, 3);
    NVIC_EnableIRQ(UART0_IRQn);
    NRF_UART0->INTENSET = UART_INTENSET_RXDRDY_Enabled << UART_INTENSET_RXDRDY_Pos;
}

int main(void)
{
    // Set 16 MHz crystal as our 16 MHz clock source (as opposed to internal RCOSC)
    NRF_CLOCK->EVENTS_HFCLKSTARTED = 0;
    NRF_CLOCK->TASKS_HFCLKSTART = 1;
    while(NRF_CLOCK->EVENTS_HFCLKSTARTED == 0)
    {
        // Wait
    }
    
    // Initialize the scheduler module
    APP_SCHED_INIT(APP_SCHED_EVENT_SIZE, APP_SCHED_QUEUE_SIZE);
    
    // Initialize fifo
    fifo_init(&s_uart_fifo);
    // Configure micro ESB
    uesb_config();
    // Configure UART
    uart_config();

    nrf_gpio_range_cfg_output(21,24);
    
    simple_uart_putstring((const uint8_t*)"\r\n Wireless UART example\n");
    
    // Start listening for incoming transmissions
    uesb_start_rx();
    
    while (true)
    {   
        app_sched_execute();
        
        // Let the CPU sleep until the next interrupt comes along
        __WFE();
        __SEV();
        __WFE();
    }
}
